package org.maltparser.core.symbol.trie;

import java.util.HashMap;

import org.maltparser.core.symbol.SymbolException;

/**

@author Johan Hall
@since 1.0
*/
public class TrieNode {
	/**
	 * Initial capacity of the hash maps.
	 */
	private final static int INITIAL_CAPACITY = 4;
	/**
	 * the character that corresponds to the trie node
	 */
	private char character;
	/**
	 * Maps a symbol table into an entry
	 */
	private HashMap<TrieSymbolTable,TrieEntry> entries;
	/**
	 * Maps a character into a child trie node
	 */
	private HashMap<Character,TrieNode> children;
	/**
	 * The parent trie node
	 */
	private TrieNode parent;
	
	/**
	 * Constructs a trie node
	 * 
	 * @param character	which character that the trie node belongs to
	 * @param parent the parent trie node
	 */
	public TrieNode(char character, TrieNode parent) {
		this.entries = new HashMap<TrieSymbolTable, TrieEntry>(INITIAL_CAPACITY);
		this.character = character;
		this.children = new HashMap<Character, TrieNode>(INITIAL_CAPACITY);
		this.parent = parent;
	}
	
	/**
	 * Adds and/or retrieve a child trie node. It only adds a entry if the parameter isWord is true.
	 * 
	 * @param isWord true if it is a word (entry), otherwise false
	 * @param c	the character to the child node
	 * @param table	which symbol table to look in or add to
	 * @param parent	the parent trie node
	 * @param code	the integer representation of the string value
	 * @return the child trie node that corresponds to the character
	 * @throws SymbolException
	 */
	public TrieNode getOrAddChild(boolean isWord, char c, TrieSymbolTable table, TrieNode parent, int code) throws SymbolException {
		TrieNode child = children.get(c);
		if (child == null) {
			child = new TrieNode(c, parent);
			children.put(c,child);
		}
		if (isWord) {
			child.addEntry(table, code);
		} 
		return child;
	}
	
	/**
	 * Adds an entry if it does not exist
	 * 
	 * @param table which symbol table to add an entry
	 * @param code the integer representation of the string value
	 * @throws SymbolException
	 */
	private void addEntry(TrieSymbolTable table, int code) throws SymbolException {
		if (table == null) {
			throw new SymbolException("Symbol table cannot be found. ");
		}
		if (!entries.containsKey(table)) {
			if (code != -1) {
				entries.put(table, new TrieEntry(code,true));
				table.updateValueCounter(code);
			} else {
				entries.put(table, new TrieEntry(table.increaseValueCounter(),false));
			}
		}
	}
	
	/**
	 * Returns the child node that corresponds to the character
	 * 
	 * @param c the character of the child node
	 * @return the child node
	 */
	public TrieNode getChild(char c) {
		return children.get(c);
	}
	
	/**
	 * Returns the entry of the symbol table 'table'
	 * 
	 * @param table	which symbol table
	 * @return the entry of the symbol table 'table'
	 */
	public TrieEntry getEntry(TrieSymbolTable table) {
		if (entries != null && table != null) {
			return entries.get(table);
		} else {
			return null;
		}
	}
	
	/**
	 * Returns the character of the trie node
	 * 
	 * @return the character of the trie node
	 */
	public char getCharacter() {
		return character;
	}
	
	/**
	 * Returns the parent node
	 * 
	 * @return the parent node
	 */
	public TrieNode getParent() {
		return parent;
	}
	
}
