package org.maltparser.parser.algorithm.covington;


import java.util.ArrayList;

import org.maltparser.core.config.SingleMaltConfiguration;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.SingleHeadedDependencyGraph;
import org.maltparser.core.graph.DependencyGraph;
import org.maltparser.core.graph.Node;
import org.maltparser.core.sentence.Sentence;
import org.maltparser.core.symbol.SymbolTable;
import org.maltparser.parser.algorithm.ParsingAlgorithm;
import org.maltparser.parser.algorithm.ParsingException;
import org.maltparser.parser.algorithm.helper.ParserAction;
import org.maltparser.parser.algorithm.helper.Transition;


/**
 * 
 * @author Joakim Nivre
 * @author Johan Hall
 * @since 1.0
*/
public abstract class Covington implements ParsingAlgorithm {
	// Behavior
	public final static int MALT_0_4  = 1;
	public final static int MALT_1_0  = 2;
	
	// Transitions
	protected static final int SHIFT = 1;
	protected static final int NOARC = 2;
	protected static final int RIGHTARC = 3;
	protected static final int LEFTARC = 4;
	
	protected Sentence sentence;
	protected DependencyGraph dependencyGraph;
	protected ParserAction parserAction;
	protected SymbolTable deprel;
	protected SingleMaltConfiguration configuration;
	
	protected boolean allowShift;
	protected int behavior;
	protected ArrayList<Node> input;
	protected int right;
	protected int left;
	protected int leftstop;
	protected int rightstop;
	
	
	public Covington(SingleMaltConfiguration configuration, Sentence sentence) throws MaltChainedException {
		this.configuration = configuration;
		this.sentence = sentence;
		this.dependencyGraph = new SingleHeadedDependencyGraph(sentence, configuration.getOptionValue("graph", "root_label").toString());
		initAllowRoot();
		initAllowShift();
		initBehavior();
		input = new ArrayList<Node>();
		initParserAction();
	}
	
	public DependencyGraph parse() throws MaltChainedException {
		clear(dependencyGraph);
		right = 1;
		while (right <= rightstop) {
			left = right -1;
			while (left >= leftstop) {
				configuration.predict();
				transition(dependencyGraph);
			}
			right++;
		}
		
		dependencyGraph.linkAllTreesToRoot();
		return dependencyGraph;
	}
	
	public DependencyGraph oracleParse(DependencyGraph goldDependencyGraph) throws MaltChainedException {
		if (!(goldDependencyGraph instanceof SingleHeadedDependencyGraph)) {
			throw new ParsingException("The gold standard graph must be a single headed graph. ");
		}
		
		clear(dependencyGraph);
		right = 1;
		while (right <= rightstop) {
			left = right - 1;
			while (left >= leftstop) { 
				oraclePredict(goldDependencyGraph);
				configuration.setInstance(parserAction);
				transition(dependencyGraph);
			}
			right++;
		}
		dependencyGraph.linkAllTreesToRoot();
		return dependencyGraph;
	}
	
	protected void transition(DependencyGraph dg) throws MaltChainedException {

		while (checkParserAction(dg) == false) {
			if (configuration.getMode() == SingleMaltConfiguration.LEARN || configuration.predictFromKBestList() == false) {
				parserAction.replaceAction(NOARC); // default parser action
				break;
			}
		}

		switch (parserAction.getLastTransition().getCode()) {
		case LEFTARC:
			dg.addArc(input.get(right), input.get(left), parserAction.getLastLabelCode());
			break;
		case RIGHTARC:
			dg.addArc(input.get(left), input.get(right), parserAction.getLastLabelCode());		
			break;
		default:
			break;
		}
		updateLeft(dg, parserAction.getLastTransition().getCode());	
	}
	
	protected boolean checkParserAction(DependencyGraph dg) throws MaltChainedException {
		int trans = parserAction.getLastTransition().getCode();
		
		if (trans == SHIFT && allowShift == false) {
			return false;
		}
		if ((trans == LEFTARC || trans == RIGHTARC) && parserAction.getLastLabelCode() == null) {
			// LabelCode cannot be found for transition LEFTARC and RIGHTARC
			return false;
		}
		/* if ((trans == SHIFT || trans == REDUCE) && parserAction.getLastLabelCode() != null) {
		    // LabelCode can be found for transition SHIFT and REDUCE 
			return false;
		}*/
		if (trans == LEFTARC && input.get(left).isRoot()) { 
			// The token on top of the stack is the root for LEFTARC
			return false;
		}
		if (trans == LEFTARC && dg.hasLabeledDependency(input.get(left))) { 
			// The token on top of the stack has already a head for transition LEFTARC
			return false;
		}
		if (trans == RIGHTARC && dg.hasLabeledDependency(input.get(right))) { 
			// The token on top of the stack has already a head for transition LEFTARC
			return false;
		}
		return true;
	}
	
	protected void oraclePredict(DependencyGraph gold) throws MaltChainedException {
		if (!input.get(left).isRoot() && gold.getTokenNode(input.get(left).getIndex()).getHead().getIndex().intValue() == input.get(right).getIndex().intValue()) {
			parserAction.addAction(LEFTARC, gold.getTokenNode(input.get(left).getIndex()).getLabelCode());
		} else if (gold.getTokenNode(input.get(right).getIndex()).getHead().getIndex().intValue() == input.get(left).getIndex().intValue()) {
			parserAction.addAction(RIGHTARC, gold.getTokenNode(input.get(right).getIndex()).getLabelCode());
		} else if (allowShift == true && (!(gold.getTokenNode(input.get(right).getIndex()).hasLeftDependent() 
				&& gold.getTokenNode(input.get(right).getIndex()).getLeftmostDependent().getIndex().intValue() < input.get(left).getIndex().intValue())
				&& !(gold.getTokenNode(input.get(right).getIndex()).getHead().getIndex().intValue() < input.get(left).getIndex().intValue() 
						&& (!gold.getTokenNode(input.get(right).getIndex()).getHead().isRoot() || leftstop == 0)))) {
			parserAction.addAction(SHIFT);
		} else {
			parserAction.addAction(NOARC);
		}
	}
	
	public boolean isOperation(String dataStructure) {
		if (dataStructure.equalsIgnoreCase("Left")) {
			return true;
		} else if (dataStructure.equalsIgnoreCase("Right")) {
			return true;
		} else if (dataStructure.equalsIgnoreCase("LeftContext")) {
			return true;
		} else if (dataStructure.equalsIgnoreCase("RightContext")) {
			return true;
		}
		return false;
	}
	
	public Node getNode(String dataStructure, int index) throws MaltChainedException {
		if (dataStructure.equals("Left")) {
			if (index < 0) {
				throw new ParsingException("Left index must be non-negative in feature specification. ");
			}
			if (behavior == Covington.MALT_0_4) {
				int tmpindex = 0;
				int tmpleft = left;
				while (tmpindex < index && tmpleft >= 0) {
					if (input.get(tmpleft).hasHead() && input.get(tmpleft).getHead().getIndex().intValue() < tmpleft) {
						tmpleft = input.get(tmpleft).getHead().getIndex();
						tmpindex++;
					} else if (input.get(tmpleft).hasHead()) {
						tmpleft--;
					} else {
						tmpleft--;
						tmpindex++;
					}
				}
				if (tmpleft >= 0) {
					return input.get(tmpleft);
				}
			} else {
				if (left-index >= 0) {
					return input.get(left-index);
				}
			}
		} else if (dataStructure.equals("Right")) {
			if (index < 0) {
				throw new ParsingException("Right index must be non-negative in feature specification. ");
			}
			if (right+index < input.size()) {
				return input.get(right+index);
			}
		} else if (dataStructure.equals("LeftContext")) {
			if (index < 0) {
				throw new ParsingException("LeftContext index must be non-negative in feature specification. ");
			}
			
			int tmpindex = 0;
			for (int i = left+1; i < right; i++) {
				if (!input.get(i).hasAncestorInside(left, right)) {
					if (tmpindex == index) {
						return input.get(i);
					} else {
						tmpindex++;
					}
				}
			}
		} else if (dataStructure.equals("RightContext")) {
			if (index < 0) {
				throw new ParsingException("RightContext index must be non-negative in feature specification. ");
			}
			int tmpindex = 0;
			for (int i = right-1; i > left; i--) {
				if (!input.get(i).hasAncestorInside(left, right)) {
					if (tmpindex == index) {
						return input.get(i);
					} else {
						tmpindex++;
					}
				}
			}
		} else {
			throw new ParsingException("Undefined data structure in feature specification. ");
		}
		return null;
	}

	private void clear(DependencyGraph dg) throws MaltChainedException {
		dg.clear();
		//dg.linkAllTokensToRoot();
		int size = dg.getSize(); //getSentenceLength()

		input.clear();
		parserAction.clear();
		for (int i = 0; i <= size; i++) {
			input.add(dg.getNode(i));
		}	
		rightstop = dg.getSize();
	}
	
	private void initAllowRoot() throws MaltChainedException {
		if ((Boolean)configuration.getOptionValue("covington", "allow_root") == true) {
			leftstop = 0;
		} else {
			leftstop = 1;
		}
	}
	
	private void initAllowShift() throws MaltChainedException {
		allowShift = (Boolean)configuration.getOptionValue("covington", "allow_shift");
	}
	
	private void initBehavior() throws MaltChainedException {
		if ((Boolean)configuration.getOptionValue("malt0.4", "behavior") == true) {
			behavior = MALT_0_4;
		} else {
			behavior = MALT_1_0;
		}		
	}
	
	protected void initParserAction() {
		Transition[] transitions = new Transition[4];

		transitions[0] = new Transition(SHIFT,"SH",false);
		transitions[1] = new Transition(NOARC,"NA",false);
		transitions[2] = new Transition(RIGHTARC,"RA",true);
		transitions[3] = new Transition(LEFTARC,"LA",true);

		parserAction = new ParserAction(transitions, this);	
	}
	
	public ParserAction getParserAction() {
		return parserAction;
	}
	
	public SingleMaltConfiguration getConfiguration() {
		return configuration;
	}

	public DependencyGraph getDependencyGraph() {
		return dependencyGraph;
	}
	
	public abstract String getName();
	protected abstract void updateLeft(DependencyGraph dg, int trans);
}


