package org.maltparser.parser.algorithm.nivre;



import org.maltparser.core.config.SingleMaltConfiguration;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.DependencyGraph;
import org.maltparser.core.sentence.Sentence;

import org.maltparser.parser.algorithm.helper.ParserAction;
import org.maltparser.parser.algorithm.helper.Transition;

/**
 * 
 * @author Joakim Nivre
 * @author Johan Hall
 * @since 1.0
*/
public class NivreEager extends Nivre {
	protected static final int REDUCE = 2;
	protected static final int RIGHTARC = 3;
	protected static final int LEFTARC = 4;
	
	public NivreEager(SingleMaltConfiguration configuration, Sentence sentence) throws MaltChainedException {
		super(configuration, sentence);
	}
	
	protected void transition(DependencyGraph dg) throws MaltChainedException {
		while (checkParserAction(dg) == false) {
			if (configuration.getMode() == SingleMaltConfiguration.LEARN || configuration.predictFromKBestList() == false) {
				parserAction.replaceAction(SHIFT); // default parser action
				break;
			}
		}

		switch (parserAction.getLastTransition().getCode()) {
		case LEFTARC:
			dg.addArc(input.peek(), stack.peek(), parserAction.getLastLabelCode());
			stack.pop();
			break;
		case RIGHTARC:
			dg.addArc(stack.peek(), input.peek(), parserAction.getLastLabelCode());
			stack.push(input.pop());
			break;
		case REDUCE:
			stack.pop();
			break;
		default:
			stack.push(input.pop()); // SHIFT
			break;
		}
	}
	
	protected boolean checkParserAction(DependencyGraph dg) throws MaltChainedException {
		int trans = parserAction.getLastTransition().getCode();
	
		if ((trans == LEFTARC || trans == RIGHTARC) && parserAction.getLastLabelCode() == null) {
			// LabelCode cannot be found for transition LEFTARC and RIGHTARC
			//System.out.println("LabelCode cannot be found for transition LEFTARC and RIGHTARC");
			return false;
		}
		/* if ((trans == SHIFT || trans == REDUCE) && parserAction.getLastLabelCode() != null) {
		    // LabelCode can be found for transition SHIFT and REDUCE 
			return false;
		}*/
		if ((trans == LEFTARC || trans == REDUCE) && stack.peek().isRoot()) { 
			// The token on top of the stack is the root for LEFTARC and REDUCE
			//System.out.println("The token on top of the stack is the root for LEFTARC and REDUCE");
			return false;
		}
		//if (trans == LEFTARC && dg.hasLabeledDependency(stack.peek())) { 
		if (trans == LEFTARC && stack.peek().hasHead()) { 
			// The token on top of the stack has already a head for transition LEFTARC
			//System.out.println("The token on top of the stack has already a head for transition LEFTARC");
			return false;
		}
		if (trans == REDUCE && !stack.peek().hasHead() && rootHandling == STRICT) {
			// The token on top of the stack have not a head for transition REDUCE
			//System.out.println("The token on top of the stack have not a head for transition REDUCE");
			return false;
		}
		return true;
	}

	
	protected void oraclePredict(DependencyGraph gold) throws MaltChainedException {
		if (!stack.peek().isRoot() && gold.getTokenNode(stack.peek().getIndex()).getHead().getIndex().intValue() == input.peek().getIndex().intValue()) {
			parserAction.addAction(LEFTARC, gold.getTokenNode(stack.peek().getIndex()).getLabelCode());
		} else if (gold.getTokenNode(input.peek().getIndex()).getHead().getIndex().intValue() == stack.peek().getIndex().intValue()) {
			parserAction.addAction(RIGHTARC, gold.getTokenNode(input.peek().getIndex()).getLabelCode());
		} else if (gold.getTokenNode(input.peek().getIndex()).hasLeftDependent() &&
				gold.getTokenNode(input.peek().getIndex()).getLeftmostDependent().getIndex().intValue() < stack.peek().getIndex().intValue()) {	
			parserAction.addAction(REDUCE);	
		/*} else if (gold.getTokenNode(input.peek().getIndex()).getHead().getIndex().intValue() < stack.peek().getIndex().intValue() && 
				(!(gold.getTokenNode(input.peek().getIndex()).getLabelCode(deprel).intValue() == gold.getRootLabels().getRootLabelCode(deprel).intValue()) )) { 
			parserAction.addAction(REDUCE); */
		} else if (gold.getTokenNode(input.peek().getIndex()).getHead().getIndex().intValue() < stack.peek().getIndex().intValue() &&
			(!gold.getTokenNode(input.peek().getIndex()).getHead().isRoot() || rootHandling == NORMAL )) {
			parserAction.addAction(REDUCE); 	
		} else {
			parserAction.addAction(SHIFT);
		}
	}

	public String getName() {
		return "nivreeager";
	}
	
	protected void initParserAction() {
		Transition[] transitions = new Transition[4];

		transitions[0] = new Transition(SHIFT,"SH",false);
		transitions[1] = new Transition(REDUCE,"RE",false);
		transitions[2] = new Transition(RIGHTARC,"RA",true);
		transitions[3] = new Transition(LEFTARC,"LA",true);

		parserAction = new ParserAction(transitions, this);	
	}
}
