package org.maltparser.parser.guide;

import java.io.File;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import org.maltparser.core.config.SingleMaltConfiguration;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.DependencyGraph;
import org.maltparser.core.sentence.Sentence;
import org.maltparser.parser.algorithm.ParsingAlgorithm;
import org.maltparser.parser.algorithm.helper.ParserAction;
import org.maltparser.parser.guide.predmodel.ModelException;
import org.maltparser.parser.guide.predmodel.PredictionModel;
/**
 * The guide is used by a parsing algorithm to predict the next parser action during parsing and to
 * add a instance to the training instance set during learning.

@author Johan Hall
@since 1.0
*/
public class Guide implements GuideModel {
	private Class<?> predictionStrategy;
	private ParsingAlgorithm parsingAlgorithm;
	private SingleMaltConfiguration configuration;
	private PredictionModel model;
	
	/**
	 * Constructs a guide.
	 * 
	 * @param configuration the current MaltParserConfiguration object
	 * @param parsingAlgorithm the parsing algorithm that uses the guide
	 * @throws MaltChainedException
	 */
	public Guide(SingleMaltConfiguration configuration, ParsingAlgorithm parsingAlgorithm) throws MaltChainedException {
		setParsingAlgorithm(parsingAlgorithm);
		setConfiguration(configuration);
		initPredictionStrategy();
		updateCardinality();
	}
	
	/**
	 * Used by the parsing algorithm for adding a instance to the training instance set during learning.
	 * 
	 * @param parserAction the current parser action.
	 * @throws MaltChainedException
	 */
	public void addInstance(ParserAction parserAction) throws MaltChainedException {
		try {
			model.addInstance(parserAction);
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ", e);
		}
	}
	
	/**
	 * Used when a sentence has been processed during learning. 
	 * 
	 * @param sentence the current sentence
	 * @param dependencyGraph the current dependency graph
	 * @throws MaltChainedException
	 */
	public void finalizeSentence(Sentence sentence, DependencyGraph dependencyGraph) throws MaltChainedException {
		try {
			model.finalizeSentence(sentence, dependencyGraph);
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ", e);
		}
	}
	
	/**
	 * Used when all training data has been processed during learning. 
	 * 
	 * @throws MaltChainedException
	 */
	public void noMoreInstances() throws MaltChainedException {
		try {
			model.noMoreInstances();
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ", e);
		}
	}
	
	/**
	 * Used for cleaning up the guide models and learner objects. Should be invoked when the
	 * system is shutting down.
	 * 
	 * @throws MaltChainedException
	 */
	public void finalize() throws MaltChainedException {
		try {
			model.finalize();
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ", e);
		}
	}
	

	/**
	 * Predicts the next parser action during parsing.
	 * 
	 * @throws MaltChainedException
	 */
	public boolean predict() throws MaltChainedException {
		try {
			return model.predict();
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ");
		}
	}

	public boolean predictFromKBestList() throws MaltChainedException {
		try {
			return model.predictFromKBestList();
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ");
		}
	}
	
	/**
	 * Updates all the features with latest cardinality (number of distinct values for a particular feature)
	 * 
	 * @throws MaltChainedException
	 */
	public void updateCardinality() throws MaltChainedException {
		try {
			model.updateCardinality();
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ");
		}
	}
	
	/**
	 * Returns a prediction strategy object (also know as a guide model)
	 * 
	 * @return a prediction strategy object (also know as a guide model)
	 */
	public Class<?> getPredictionStrategy() {
		return predictionStrategy;
	}

	/**
	 * Initialize prediction strategy (also know as a guide model)
	 * 
	 * @throws MaltChainedException
	 */
	public void initPredictionStrategy() throws MaltChainedException {
		predictionStrategy = (Class<?>)getConfiguration().getOptionValue("guide", "prediction_strategy");
		Class<?>[] argTypes = { org.maltparser.parser.guide.Guide.class };
		
		Object[] arguments = new Object[1];
		arguments[0] = this;

		try {	
			Constructor<?> constructor = predictionStrategy.getConstructor(argTypes);
			model = (PredictionModel)constructor.newInstance(arguments);
		} catch (NoSuchMethodException e) {
			throw new ModelException("The guide model class '"+predictionStrategy.getName()+"' cannot be initialized. ", e);
		} catch (InstantiationException e) {
			throw new ModelException("The guide model class '"+predictionStrategy.getName()+"' cannot be initialized. ", e);
		} catch (IllegalAccessException e) {
			throw new ModelException("The guide model class '"+predictionStrategy.getName()+"' cannot be initialized. ", e);
		} catch (InvocationTargetException e) {
			throw new ModelException("The guide model class '"+predictionStrategy.getName()+"' cannot be initialized. ", e);
		}
	}

	/**
	 * Returns the parsing algorithm 
	 * 
	 * @return the parsing algorithm 
	 */
	public ParsingAlgorithm getParsingAlgorithm() {
		return parsingAlgorithm;
	}

	/**
	 * Sets the parsing algorithm that use the guide to predict the next parser action during parsing 
	 * and to add a instance during learning.
	 * 
	 * @param parsingAlgorithm the parsing algorithm that use the guide
	 */
	public void setParsingAlgorithm(ParsingAlgorithm parsingAlgorithm) {
		this.parsingAlgorithm = parsingAlgorithm;
	}

	/**
	 * Returns the current MaltParser configuration object
	 * 
	 * @return the current MaltParser configuration object
	 */
	public SingleMaltConfiguration getConfiguration() {
		return configuration;
	}

	/**
	 * Sets the current MaltParser configuration object
	 * 
	 * @param configuration the current MaltParser configuration object
	 */
	public void setConfiguration(SingleMaltConfiguration configuration) {
		this.configuration = configuration;
	}
	
	public String getModelName() throws MaltChainedException {
		try {
			return model.getModelName();
		} catch (NullPointerException e) {
			throw new GuideException("The guide model cannot be found. ");
		}
	}
	
	/* public String getPath() throws MaltChainedException {
		if (configuration != null) {
			return configuration.getPath().getPath();
		}
		return null;
	}*/
	
	public OutputStreamWriter getOutputStreamWriter(String suffix) throws MaltChainedException {
		if (configuration != null) {
			return configuration.getOutputStreamWriter(suffix);
		}
		return null;
	}
	
	public InputStreamReader getInputStreamReader(String suffix) throws MaltChainedException {
		if (configuration != null) {
			return configuration.getInputStreamReader(suffix);
		}
		return null;
	}
	
	public File getFile(String suffix) throws MaltChainedException {
		if (configuration != null) {
			return configuration.getFile(suffix);
		}
		return null;
	}
	
	public int getParserMode() throws MaltChainedException {
		if (configuration != null) {
			return configuration.getMode();
		}
		return -1;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		//TODO
		return sb.toString();
	}
}
