package org.maltparser.parser.guide.feature;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.sentence.Token;
import org.maltparser.core.symbol.SymbolTable;
import org.maltparser.core.symbol.nullvalue.NullValues.NullValueId;
import org.maltparser.parser.algorithm.ParsingAlgorithm;
/**


@author Johan Hall
@since 1.0
 */
public class SuffixFeature extends Feature {
	private InputColumnFeature feature;
	private SymbolTable table;
	private Integer suffixLength;

	public SuffixFeature(ParsingAlgorithm parsingAlgorithm, String spec) throws MaltChainedException {
		super();
		super.initialize(parsingAlgorithm);
		spec = spec.trim();
		int index = spec.lastIndexOf(',');
		if (index == -1) {
			throw new FeatureException("Suffix feature '"+spec+"' has not the correct format. ");
		}
		if (spec.substring(0,index).trim().startsWith("Suffix")) {
			throw new FeatureException("Could not have recursive suffix feature at line '"+spec+"'. ");
		}
		
		Feature f = initParentFeature(spec.substring(0,index).trim());
		if (!init(spec.substring(index+1))) { 
			throw new FeatureException("Suffix feature could not be initialized at line '"+spec+"'. ");
		}
		if (f instanceof InputColumnFeature) {
			feature = (InputColumnFeature)f;
			table = parsingAlgorithm.getConfiguration().getSymbolTables().addSymbolTable("SUFFIX_"+suffixLength+"_"+feature.getColumn().getName(), feature.getSymbolTable());
		} else {
			throw new FeatureException("It is only possible to use a suffix feature on input column feature. ");
		}
	}
	
	protected Feature initParentFeature(String spec) throws MaltChainedException {
		spec = spec.trim();
		 int index = spec.indexOf('(');
		 if (index == -1) {
			 throw new FeatureException("There is no start parenthesis at line '"+spec+"' in the feature specification file. ");
		 }
		 if (!spec.endsWith(")")) {
			 throw new FeatureException("There is no end parenthesis at line '"+spec+"' in the feature specification file. ");
		 }
		 Class<?> clazz;
		 try {
			clazz = Class.forName("org.maltparser.parser.guide.feature."+spec.substring(0, index)+"Feature");
		 } catch (ClassNotFoundException e) {
			 throw new FeatureException("The feature class '"+spec.substring(0, index)+"' cannot be found. ", e);
		 }
		 try {
			Class<?>[] argTypes = { org.maltparser.parser.algorithm.ParsingAlgorithm.class, java.lang.String.class };
			Object[] arguments = new Object[2];
			arguments[0] = parsingAlgorithm;
			arguments[1] = spec.substring(index+1,spec.length()-1);
			Constructor<?> constructor = clazz.getConstructor(argTypes);
			return (Feature)constructor.newInstance(arguments);
		} catch (NoSuchMethodException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InstantiationException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (IllegalAccessException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InvocationTargetException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		}
	}
	
	public boolean init(String spec) throws MaltChainedException {
		if (spec.trim().length() >= 1) {
			try {
				suffixLength = Integer.parseInt(spec.trim());
			} catch (NumberFormatException e) {
				throw new FeatureException("Suffix length contain a non-integer value. ", e);
			}
			return true;
		}
		return false;
	}
	
	public Integer getIntegerValue(String value) throws MaltChainedException {
		return table.getSymbolStringToCode(value);
	}

	public String getStringValue(Integer value) throws MaltChainedException {
		return table.getSymbolCodeToString(value);
	}

	public void setValue(Integer value) throws MaltChainedException {
		featureValues.clear();
		featureKnown.clear();
		if (table.getSymbolCodeToString(value) == null) {
			featureKnown.put(value, table.getKnown(value));
			featureValues.put(value, table.getNullValueSymbol(NullValueId.NO_NODE));
			nullValue = true;
		} else {
			featureKnown.put(value, table.getKnown(value));
			featureValues.put(value, table.getSymbolCodeToString(value));
			nullValue = table.isNullValue(value);
		}
	}

	public void setValue(String value) throws MaltChainedException {
		featureValues.clear();
		featureKnown.clear();
		if (table.getSymbolStringToCode(value) < 0) {
			featureKnown.put(table.getNullValueCode(NullValueId.NO_NODE), table.getKnown(value));
			featureValues.put(table.getNullValueCode(NullValueId.NO_NODE), value);
			nullValue = true;
		} else {
			featureKnown.put(table.getSymbolStringToCode(value), table.getKnown(value));
			featureValues.put(table.getSymbolStringToCode(value), value);
			nullValue = table.isNullValue(value);
		}
	}

	public void update() throws MaltChainedException {
		feature.update();
		featureValues.clear();
		featureKnown.clear();
		if (feature.getCurrentUpdateNode() != null) {
			Token token = feature.getCurrentUpdateNode().getToken();
			if (token != null) {
				String str = token.getColumnInstance(feature.getColumn()).getStringValue();
				String suffixStr;
				if (str.length()-suffixLength > 0) {
					suffixStr = str.substring(str.length()-suffixLength);
				} else {
					suffixStr = str;
				}
				int code = table.addSymbol(suffixStr);
				featureValues.put(code, suffixStr); 
				featureKnown.put(code, table.getKnown(suffixStr));
				nullValue = false;
			} else {
				featureKnown.put(table.getNullValueCode(NullValueId.ROOT_NODE), true);
				featureValues.put(table.getNullValueCode(NullValueId.ROOT_NODE), table.getNullValueSymbol(NullValueId.ROOT_NODE));
				nullValue = true;
			}
		} else {
			featureValues.put(table.getNullValueCode(NullValueId.NO_NODE), table.getNullValueSymbol(NullValueId.NO_NODE));
			featureKnown.put(table.getNullValueCode(NullValueId.NO_NODE), true);
			nullValue = true;
		}
	}

	public Feature getParentFeature() {
		return feature;
	}
	
	public void updateCardinality() throws MaltChainedException {
		super.setCardinality(table.getValueCounter()); 
		
	}
	
	public String toString() {
		return "Suffix("+feature.toString()+", "+suffixLength+")";
	}	
}
